/*PROGRAM NAME: WIPER CONTROLLER

    DESCRIPTION : This is the program for Arduino nano to control the soap wiper and dropper devices of the
                  SoapFilmScope.
                  The wiper device is used to wide the SoapFilm tube opening with a sponge soaked with the soap solution
                  to reform the soap film.
 
                  The device is controlled by an Arduino nano using this program.
                  The joystick module is used to drive a menu on a LCD 20x2, set up the parameters used by the wiper.

                  The Y-positions of the joystick is used to select the menu option, the joystick
                  button to select/confirm and the X-positions to assign values to the different menu options.

  LCD MENU DESCRIPTION:


                  1. Wipe Soap Film:      Regenerate the soap film.
                  2. Set Wiper Angles:    Set the wiper movement.
                  3. Set Wiper Speed:     Set the wiper movement.                 
  }

    AUTHOR : Danilo Roccatano (c)
    VERSION: 1.2
    DATE   : Novemeber 2024
*/

#include <Servo.h>
#include <EEPROM.h>
#include <Wire.h>
#include <LiquidCrystal_I2C.h>

// Function prototypes
void DisplayMenu(int );
void lcdMenuDisplay(int , int );
void clearLines(int );
void DisplayLCDVar(int , int , int row);
void adjustServoAngles();
void adjustServoWipeSpeed();

// LCD settings
const uint8_t I2C_ADDRESS = 0x27;
const uint8_t LCD_CHAR = 16;
const uint8_t LCD_LINE = 2;
LiquidCrystal_I2C lcd(I2C_ADDRESS, LCD_CHAR, LCD_LINE);

// Joystick setup
#define XJOY_PIN A1
#define YJOY_PIN A0
#define BUTTON_PIN 5

int buttonState = 0;
int imenu = 0;  // Select menu item
bool wiperAtUpper = true;  // Track wiper position; starts at the upper angle

// Servo motor setup (for the wiper)
Servo wiperServo;
int wiperUpperAngle, wiperLowerAngle; // Values to be stored in EEPROM
int wiperPosition;

// Wipe speed setting
int wipeSpeed = 10; // Default wipe speed delay in milliseconds (1000 ms = 1 second)

// LCD menu titles
const char TITLE_LCD[5][20] = {
  "SFS Controller",
  "v1.0 D.Roccatano",
  "X-ANGLE0: ",
  "Y-ANGLE1: ",
  "Set Wiper Speed:"
};
// Menu titles
char MENU_TITLE[3][20] = {
  "1. Wipe Soap Film",
  "2. Set Wip.Angles",
  "3. Set Wip.Speed"
};

void setup() {
    Serial.begin(9600);

    // Initialize button pin
    pinMode(BUTTON_PIN, INPUT_PULLUP);

    // Initialize LCD
    lcd.init();
    lcd.backlight();
    lcd.setCursor(0, 0);
    lcd.print(TITLE_LCD[0]);
    lcd.setCursor(0, 1);
    lcd.print(TITLE_LCD[1]);
    delay(2000);

    // Initialize servo motor
    wiperServo.attach(12);  // Pin for wiper servo

    // Retrieve the saved angles and speed from EEPROM
    EEPROM.get(0, wiperUpperAngle);      // Load `wiperUpperAngle` from address 0
    EEPROM.get(2, wiperLowerAngle);      // Load `wiperLowerAngle` from address 2
    EEPROM.get(4, wipeSpeed);            // Load `wipeSpeed` from address 4
    EEPROM.get(6, wiperPosition);        // Load `wiperPosition` from address 6

    // Set default values if none were stored (assuming uninitialized EEPROM returns 255 or invalid values)
    if (wiperUpperAngle == 255 || wiperUpperAngle > 180) wiperUpperAngle = 180;
    if (wiperLowerAngle == 255 || wiperLowerAngle < 0) wiperLowerAngle = 0;
    if (wipeSpeed == 255 || wipeSpeed <= 0) wipeSpeed = 10; // Default speed if invalid
    if (wiperPosition < wiperLowerAngle || wiperPosition > wiperUpperAngle) {
        wiperPosition = wiperUpperAngle;  // Set to upper angle if out of bounds
    }

    // Set servo to the initial position
    wiperServo.write(wiperPosition);

    delay(2000);
    DisplayMenu(0);  // Display initial menu
}
void loop() {
  buttonState = digitalRead(BUTTON_PIN);

  if (!buttonState) {
    switch (imenu) {
      // Handle menu selection
      case 0: // Form the soap film (wiper)
        RunWiper();
        break;      
      case 1: // Set wiper angles
        adjustServoAngles();
        break;
      case 2: // Set wiper speed
        adjustServoWipeSpeed();
        break;
    }
    delay(300);  // Debounce delay
    DisplayMenu(imenu);
  }

  // Handle joystick movement
  int joystickYVal = analogRead(YJOY_PIN);
  if (joystickYVal >= 550) {
    imenu = (imenu + 1) % 3;
    DisplayMenu(imenu);
    delay(200);
  } else if (joystickYVal <= 300) {
    imenu--;
    if (imenu < 0) imenu = 2;
    DisplayMenu(imenu);
    delay(200);
  }
}

//
// FUNCTIONS
//

// Function to display the current menu
void DisplayMenu(int imenu) {
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(MENU_TITLE[imenu]);
}

void RunWiper() {
  if (wiperAtUpper) {
    // Move from wiperUpperAngle to wiperLowerAngle step by step
    for (int pos = wiperUpperAngle; pos >= wiperLowerAngle; pos--) {
      wiperServo.write(pos);
      delay(wipeSpeed);  // Delay between each small movement step
    }
    wiperAtUpper = false;  // Update the flag to indicate wiper is now at the lower angle
  } else {
    // Move from wiperLowerAngle to wiperUpperAngle step by step
    for (int pos = wiperLowerAngle; pos <= wiperUpperAngle; pos++) {
      wiperServo.write(pos);
      delay(wipeSpeed);  // Delay between each small movement step
    }
    wiperAtUpper = true;  // Update the flag to indicate wiper is now at the upper angle
  }
}

void DisplayLCDVar(int pos, int val, int row)
{
  lcd.setCursor(pos, row);
  int numSpaces = 16 - pos;
  for (int i = 0; i < numSpaces; i++) {
    lcd.print(" ");
  }
  lcd.setCursor(pos, row);
  lcd.print(val);
}

void clearLines(int row) {
  for (int i = 0; i < LCD_CHAR; i++) {
    if (row < 2) {
      lcd.setCursor(i, row);
      lcd.write(254);
    } else {
      lcd.setCursor(i, 0);
      lcd.write(254);
      lcd.setCursor(i, 1);
      lcd.write(254);
    }
  }
}

void adjustServoAngles()
{
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(TITLE_LCD[2]);
  lcd.setCursor(0, 1);
  lcd.print(TITLE_LCD[3]);
  DisplayLCDVar(10, wiperLowerAngle, 0);
  DisplayLCDVar(10, wiperUpperAngle, 1);
  delay(200);

  int buttonState = digitalRead(BUTTON_PIN);

  while (buttonState) {
    int joystickXVal = analogRead(XJOY_PIN);
    int joystickYVal = analogRead(YJOY_PIN);

    while (joystickXVal <= 300) {
      wiperLowerAngle++;
      if (wiperLowerAngle > 180)
        wiperLowerAngle = 0;
      DisplayLCDVar(10, wiperLowerAngle, 0);
      delay(200);
      joystickXVal = analogRead(XJOY_PIN);
      wiperServo.write(wiperLowerAngle);
    }
    while (joystickXVal >= 550) {
      wiperLowerAngle--;
      if (wiperLowerAngle < 0)
        wiperLowerAngle = 180;
      DisplayLCDVar(10, wiperLowerAngle, 0);
      delay(200);
      joystickXVal = analogRead(XJOY_PIN);
      wiperServo.write(wiperLowerAngle);
    }
    while (joystickYVal >= 550) {
      wiperUpperAngle--;
      if (wiperUpperAngle < 0)
        wiperUpperAngle = 180;
      DisplayLCDVar(10, wiperUpperAngle, 1);
      delay(200);
      joystickYVal = analogRead(YJOY_PIN);
      wiperServo.write(wiperUpperAngle);
    }
    while (joystickYVal <= 300) {
      wiperUpperAngle++;
      if (wiperUpperAngle > 180)
        wiperUpperAngle = 0;
      DisplayLCDVar(10, wiperUpperAngle, 1);
      delay(200);
      joystickYVal = analogRead(YJOY_PIN);
      wiperServo.write(wiperUpperAngle);
    }
    buttonState = digitalRead(BUTTON_PIN); // Update buttonState inside the loop
  }
  clearLines(2);
  // Store updated angles in EEPROM
   EEPROM.put(0, wiperUpperAngle);              // Store `wiperUpperAngle` at address 0
   EEPROM.put(2, wiperLowerAngle);              // Store `wiperLowerAngle` at address 2
}

// Adjust servo wipe speed
void adjustServoWipeSpeed() {
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(TITLE_LCD[4]);
  DisplayLCDVar(0, wipeSpeed, 1); // Display current speed
  delay(200);

  int buttonState = digitalRead(BUTTON_PIN);

  while (buttonState) {
    int joystickXVal = analogRead(XJOY_PIN);    
    int joystickYVal = analogRead(YJOY_PIN);

    if (joystickXVal >= 550) { // Increase speed
      wipeSpeed += 1;
      if (wipeSpeed > 1000) wipeSpeed = 1000; // Set max limit for speed
      DisplayLCDVar(0, wipeSpeed, 1);
      delay(200);
    } else if (joystickXVal <= 300) { // Decrease speed
      wipeSpeed -= 1;
      if (wipeSpeed < 1) wipeSpeed = 1; // Set min limit for speed
      DisplayLCDVar(0, wipeSpeed, 1);
      delay(200);
    }
    
    if (joystickYVal >= 550) { // Increase speed
      wipeSpeed += 10;
      if (wipeSpeed > 1000) wipeSpeed = 1000; // Set max limit for speed
      DisplayLCDVar(0, wipeSpeed, 1);
      delay(200);
    } else if (joystickYVal <= 300) { // Decrease speed
      wipeSpeed -= 10;
      if (wipeSpeed < 1) wipeSpeed = 1; // Set min limit for speed
      DisplayLCDVar(0, wipeSpeed, 1);
      delay(200);
    }

    buttonState = digitalRead(BUTTON_PIN); // Update button state to exit
  }
    // Store updated angles in EEPROM
   EEPROM.put(4, wipeSpeed);                    // Store `wipeSpeed` at address 4
   EEPROM.put(6, wiperPosition);                // Store `wiperPosition` at address 6
}
